/*
** Copyright (C) 2001-2010 by Carnegie Mellon University.
**
** @OPENSOURCE_HEADER_START@
**
** Use of the SILK system and related source code is subject to the terms
** of the following licenses:
**
** GNU Public License (GPL) Rights pursuant to Version 2, June 1991
** Government Purpose License Rights (GPLR) pursuant to DFARS 252.227.7013
**
** NO WARRANTY
**
** ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
** PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
** PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
** "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
** KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
** LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
** MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
** OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
** SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
** TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
** WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
** LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
** CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
** CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
** DELIVERABLES UNDER THIS LICENSE.
**
** Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
** Mellon University, its trustees, officers, employees, and agents from
** all claims or demands made against them (and any related losses,
** expenses, or attorney's fees) arising out of, or relating to Licensee's
** and/or its sub licensees' negligent use or willful misuse of or
** negligent conduct or willful misconduct regarding the Software,
** facilities, or other rights or assistance granted by Carnegie Mellon
** University under this License, including, but not limited to, any
** claims of product liability, personal injury, death, damage to
** property, or violation of any laws or regulations.
**
** Carnegie Mellon University Software Engineering Institute authored
** documents are sponsored by the U.S. Department of Defense under
** Contract F19628-00-C-0003. Carnegie Mellon University retains
** copyrights in all material produced under this contract. The U.S.
** Government retains a non-exclusive, royalty-free license to publish or
** reproduce these documents, or allow others to do so, for U.S.
** Government purposes only pursuant to the copyright license under the
** contract clause at 252.227.7013.
**
** @OPENSOURCE_HEADER_END@
*/

/*
**  rwstatssetup.c
**
**  Application setup for rwstats.  See rwstats.c for a description.
*/

#include <silk/silk.h>

RCSIDENT("$SiLK: rwstatssetup.c 15752 2010-05-07 21:24:38Z mthomas $");

#include "rwstats.h"
#include "rwuniqtemp.h"
#include <silk/sksite.h>
#include <silk/iochecks.h>


/* TYPEDEFS AND DEFINES */

/* file handle for --help usage message */
#define USAGE_FH stdout

/* where to write filenames if --print-file specified */
#define PRINT_FILENAMES_FH  stderr


/* LOCAL VARIABLES */

/* did user ask for legacy help? */
static int legacy_help_requested = 0;

/* the number of bytes that the disinct counts put into the hash key
 * and into the heap value */
static size_t distinct_bytes_key = 0;
static size_t distinct_bytes_value = 0;

/* key fields used when parsing the user's --fields switch */
static sk_stringmap_t *key_field_map = NULL;

/* the text the user entered for the --fields switch */
static const char *fields_arg = NULL;

/* value fields used when parsing the user's --values switch */
static sk_stringmap_t *value_field_map = NULL;

/* the text the user entered for the --values switch */
static const char *values_arg = NULL;

/* information about each potential "value" field the user can choose
 * to compute and display. */
static value_field_t potential_values[] = {
    {"Bytes",         sizeof(uint64_t), 0, 20, VALUE_BYTES,   NULL},
    {"Packets",       sizeof(uint32_t), 0, 10, VALUE_PACKETS, NULL},
    {"Records",       sizeof(uint32_t), 0, 10, VALUE_FLOWS,   NULL},
    {"sIP-Distinct",  sizeof(uint64_t), 0, 10, DISTINCT_SIP,  NULL},
    {"dIP-Distinct",  sizeof(uint64_t), 0, 10, DISTINCT_DIP,  NULL}
};

/* name of program to run to page output */
static char *pager;

/* the floor of the sTime and/or eTime */
static uint32_t time_bin_size = 0;

/* when time-binning is active and all time fields---sTime, eTime,
 * elapased---are requested, we must adjust 'elapsed' so that it is
 * equal to eTime-sTime.  this is non-0 if we must adjust */
static int adjust_elapsed = 0;

#define PARSE_KEY_ELAPSED   (1 << 0)
#define PARSE_KEY_STIME     (1 << 1)
#define PARSE_KEY_ETIME     (1 << 2)

/* which of elapsed, sTime, and eTime were requested. uses above values */
static unsigned int time_fields = 0;

/* input checker */
static iochecksInfoStruct_t *ioISP = NULL;

/* plug-ins to attempt to load at startup */
static const char *app_plugin_names[] = {
    SK_PLUGIN_ADD_SUFFIX("addrtype"),
    SK_PLUGIN_ADD_SUFFIX("ccfilter"),
    SK_PLUGIN_ADD_SUFFIX("pmapfilter"),
    SK_PLUGIN_ADD_SUFFIX("silkpython"),
    NULL /* sentinel */
};

/* non-zero if we are shutting down due to a signal; controls whether
 * errors are printed in appTeardown(). */
static int caught_signal = 0;


/* OPTIONS */

static struct option appOptions[] = {
    {"overall-stats",       NO_ARG,       0, OPT_OVERALL_STATS},
    {"detail-proto-stats",  REQUIRED_ARG, 0, OPT_DETAIL_PROTO_STATS},

    {"fields",              REQUIRED_ARG, 0, OPT_FIELDS},
    {"values",              REQUIRED_ARG, 0, OPT_VALUES},
    {"plugin",              REQUIRED_ARG, 0, OPT_PLUGIN},

    {"count",               REQUIRED_ARG, 0, OPT_COUNT},
    {"threshold",           REQUIRED_ARG, 0, OPT_THRESHOLD},
    {"percentage",          REQUIRED_ARG, 0, OPT_PERCENTAGE},

    {"top",                 NO_ARG,       0, OPT_TOP},
    {"bottom",              NO_ARG,       0, OPT_BOTTOM},

    {"presorted-input",     NO_ARG,       0, OPT_PRESORTED_INPUT},
    {"no-percents",         NO_ARG,       0, OPT_NO_PERCENTS},
    {"bin-time",            OPTIONAL_ARG, 0, OPT_BIN_TIME},
    {"epoch-time",          NO_ARG,       0, OPT_EPOCH_TIME},
    {"integer-ips",         NO_ARG,       0, OPT_INTEGER_IPS},
    {"zero-pad-ips",        NO_ARG,       0, OPT_ZERO_PAD_IPS},
    {"integer-sensors",     NO_ARG,       0, OPT_INTEGER_SENSORS},
    {"no-titles",           NO_ARG,       0, OPT_NO_TITLES},
    {"no-columns",          NO_ARG,       0, OPT_NO_COLUMNS},
    {"column-separator",    REQUIRED_ARG, 0, OPT_COLUMN_SEPARATOR},
    {"no-final-delimiter",  NO_ARG,       0, OPT_NO_FINAL_DELIMITER},
    {"delimited",           OPTIONAL_ARG, 0, OPT_DELIMITED},
    {"print-filenames",     NO_ARG,       0, OPT_PRINT_FILENAMES},
    {"copy-input",          REQUIRED_ARG, 0, OPT_COPY_INPUT},
    {"output-path",         REQUIRED_ARG, 0, OPT_OUTPUT_PATH},
    {"pager",               REQUIRED_ARG, 0, OPT_PAGER},
    {"temp-directory",      REQUIRED_ARG, 0, OPT_TEMP_DIRECTORY},
    {"legacy-timestamps",   OPTIONAL_ARG, 0, OPT_LEGACY_TIMESTAMPS},

    {"legacy-help",         NO_ARG,       0, OPT_LEGACY_HELP},

    {0,0,0,0}               /* sentinel entry */
};

static const char *appHelp[] = {
    ("Print minima, maxima, quartiles, and interval-count\n"
     "\tstatistics for bytes, pkts, bytes/pkt across all flows.  Def. No"),
    ("Print above statistics for each of the specified\n"
     "\tprotocols.  List protocols or ranges separated by commas. Def. No"),

    NULL, /* generated dynamically */
    NULL, /* generated dynamically */
    ("Load given plug-in to add fields and/or values. Switch may\n"
     "\tbe repeated to load multiple plug-ins. Def. None"),

    ("Print the specified number of bins"),
    ("Print bins where the primary value is greater-/less-than\n"
     "\tthis threshold. Not valid for primary values from plug-ins."),
    ("Print bins where the primary value is greater-/less-than\n"
     "\tthis percentage of the total across all flows. Only allowed when the\n"
     "\tprimary value field is Bytes, Packets, or Records."),

    ("Print the top N keys and their values. Def. Yes"),
    ("Print the bottom N keys and their values. Def. No"),

    ("Assume input has been presorted using\n"
     "\trwsort invoked with the exact same --fields value. Def. No"),
    ("Do not print the percentage columns. Def. Print percents"),
    ("When using 'sTime' or 'eTime' as a key, adjust time(s) to\n"
     "\tto appear in N-second bins (floor of time is used). Def. No, "),
    "Print times in UNIX epoch seconds. Def. No",
    "Print IP numbers as integers. Def. Canonical form",
    "Print IP numbers in zero-padded canonical form. Def. No",
    "Print sensor as an integer. Def. Sensor name",
    "Do not print column titles. Def. Print titles",
    "Disable fixed-width columnar output. Def. Columnar",
    "Use specified character between columns. Def. '|'",
    "Suppress column delimiter at end of line. Def. No",
    "Shortcut for --no-columns --no-final-del --column-sep=CHAR",
    "Print names of input files as they are opened. Def. No",
    "Copy all input SiLK Flows to given pipe or file. Def. No",
    "Send output to given file path. Def. stdout",
    "Program to invoke to page output. Def. $SILK_PAGER or $PAGER",
    NULL, /* generated dynamically */
    ("Timestamp format. Choices:\n"
     "\t0==(new)\"YYYY/MM/DDThh:mm:ss\"; 1==(legacy)\"MM/DD/YYYY hh:mm:ss\".\n"
     "\tDef. 0 when switch not provided; 1 when switch provided with no value"),
    "Print help, including legacy switches",
    (char *)NULL
};



/* LOCAL FUNCTION PROTOTYPES */

static void appUsageLong(void);
static void appHandleSignal(int sig);

static int createStringmaps(void);
static int parseKeyFields(const char *field_string);
static int parseValueFields(const char *field_string);
static void topnSetup(const rwstats_legacy_t *leg);


/* Adding key and value fields */
static key_field_t *addKeyField(
    uint32_t        id,
    size_t          byte_width,
    rec2key_fn_t    rec2key,
    key2rec_fn_t    key2rec);

static int appAddPluginKey(
    sk_stringmap_entry_t   *map_entry);


/* Sometimes key fields must be modified---e.g., sTime and eTime when
 * the --bin-time switch is given.  These functions do the necessary
 * modification. */
static void recToKeyStime(
    const rwRec    *rwrec,
    void           *bin_value);
static void recToKeyEtime(
    const rwRec    *rwrec,
    void           *bin_value);
static void recToKeyElapsed(
    const rwRec    *rwrec,
    void           *bin_value);
static void keyToRecStime(
    rwRec          *rwrec,
    const void     *bin_value);
static void keyToRecEtime(
    rwRec          *rwrec,
    const void     *bin_value);
static void keyToRecElapsed(
    rwRec          *rwrec,
    const void     *bin_value);


/* FUNCTION DEFINITIONS */

/*
 *  appUsageLong();
 *
 *    Print complete usage information to USAGE_FH.  Pass this
 *    function to skOptionsSetUsageCallback(); skOptionsParse() will
 *    call this funciton and then exit the program when the --help
 *    option is given.
 */
static void appUsageLong(void)
{
    FILE *fh = USAGE_FH;
    int i;

    /* use two macros to avoid CPP limit in c89 */
#define USAGE_MSG                                                             \
    ("<SWITCHES> [FILES]\n"                                                   \
     "\tSummarize SiLK Flow records by the specified field(s) into bins.\n"   \
     "\tFor each bin, compute the specified value(s), then display the\n"     \
     "\tresults as a Top-N or Bottom-N list based on the primary value.\n"    \
     "\tThe N can be a fixed value; some values allow the N to be a\n"        \
     "\tthreshold value or to be based on a percentage of the input.\n")
#define USAGE_MSG_2                                                           \
    ("\tAlternatively, provide statistics for each of bytes, packets, and\n" \
     "\tbytes-per-packet giving minima, maxima, quartile, and interval\n"     \
     "\tflow-counts across all flows or across user-specified protocols.\n"   \
     "\tWhen no files are given on command line, flows are read from STDIN.\n")

    /* set up the --fields and --values */
    createStringmaps();

    fprintf(fh, "%s %s%s", skAppName(), USAGE_MSG, USAGE_MSG_2);

    for (i = 0; appOptions[i].name; i++) {
        /* Print a header before some options */
        switch ((appOptionsEnum)i) {
          case OPT_OVERALL_STATS:
            fprintf(fh, "\nPROTOCOL STATISTICS SWITCHES:\n");
            break;
          case OPT_FIELDS:
            fprintf(fh, "\nTOP-N/BOTTOM-N SWITCHES:\n");
            break;
          case OPT_COUNT:
            fprintf(fh, ("\nHow to determine the N for Top-/Bottom-N;"
                         " must specify one:\n"));
            break;
          case OPT_TOP:
            fprintf(fh, ("\nWhether to compute Top- or Bottom-N;"
                         " may specify one (top is default):\n"));
            break;
          case OPT_PRESORTED_INPUT:
            fprintf(fh, "\nMISCELLANEOUS SWITCHES:\n");
            skOptionsDefaultUsage(fh);
            break;
          case OPT_BIN_TIME:
            skIPv6PolicyUsage(fh);
            break;
          default:
            break;
        }

        fprintf(fh, "--%s %s. ", appOptions[i].name,
                SK_OPTION_HAS_ARG(appOptions[i]));
        switch ((appOptionsEnum)i) {
          case OPT_FIELDS:
            /* Dynamically build the help */
            fprintf(fh, "Field(s) to use as key as a comma separated list:\n");
            skStringMapPrintUsage(key_field_map, fh, 4);
            break;
          case OPT_VALUES:
            fprintf(fh, ("Value(s) to compute, a comma separated list."
                         " Def. flows.\n"
                         "\tThe first value will be used as the"
                         " basis for the Top-N/Bottom-N.\n"));
            skStringMapPrintUsage(value_field_map, fh, 4);
            break;
          case OPT_BIN_TIME:
            fprintf(fh, "%s%d\n", appHelp[i], DEFAULT_TIME_BIN);
            break;
          case OPT_TEMP_DIRECTORY:
            fprintf(fh,
                    ("Store temporary files in this directory.\n"
                     "\tDef. $" TMPDIR_ENVAR1 " or $" TMPDIR_ENVAR2));
#ifdef DEFAULT_TEMP_DIR
            fprintf(fh, " or %s", DEFAULT_TEMP_DIR);
#endif
            fprintf(fh, "\n");
            break;
          default:
            /* Simple help text from the appHelp array */
            fprintf(fh, "%s\n", appHelp[i]);
            break;
        }
    }

    sksiteOptionsUsage(fh);
    skPluginOptionsUsage(fh);
    if (legacy_help_requested) {
        legacyOptionsUsage(fh);
    }
}


/*
 *  appSetup(argc, argv);
 *
 *    Perform all the setup for this application include setting up
 *    required modules, parsing options, etc.  This function should be
 *    passed the same arguments that were passed into main().
 *
 *    Returns to the caller if all setup succeeds.  If anything fails,
 *    this function will cause the application to exit with a FAILURE
 *    exit status.
 */
void appSetup(int argc, char **argv)
{
    rwstats_legacy_t leg;
    int using_pager = 0;
    int j;

    /* verify same number of options and help strings */
    assert((sizeof(appHelp)/sizeof(char *)) ==
           (sizeof(appOptions)/sizeof(struct option)));

#ifdef DEFAULT_TEMP_DIR
    assert(DEFAULT_TEMP_DIR != NULL);
#endif

    /* register the application */
    skAppRegister(argv[0]);
    skOptionsSetUsageCallback(&appUsageLong);

    /* initialize globals */
    output.pathname = "stdout";
    output.fp = stdout;
    value_byte_width = 0;
    value_num_fields = 0;
    memset(&app_flags, 0, sizeof(app_flags));
    memset(&leg, 0, sizeof(rwstats_legacy_t));

    /* initialize plugin library */
    skPluginSetup(2, SKPLUGIN_APP_STATS_FIELD, SKPLUGIN_APP_STATS_VALUE);

    /* allow for 0 input and 0 output pipes */
    ioISP = iochecksSetup(1, 0, argc, argv);

    /* register the options */
    if (skOptionsRegister(appOptions, &appOptionsHandler, NULL)
        || legacyOptionsSetup(&leg)
        || sksiteOptionsRegister(SK_SITE_FLAG_CONFIG_FILE)
        || skIPv6PolicyOptionsRegister(&ipv6_policy))
    {
        skAppPrintErr("Unable to register options");
        appExit(EXIT_FAILURE);
    }

    /* try to load run-time plugins */
    for (j = 0; app_plugin_names[j]; ++j) {
        skPluginLoadPluginOrDynlib(app_plugin_names[j], DYNLIB_UNIQ, 0);
    }

    /* parse options */
    ioISP->firstFile = skOptionsParse(argc, argv);
    if (ioISP->firstFile < 0) {
        skAppUsage();/* never returns */
    }

    /* try to load site config file; if it fails, we will not be able
     * to resolve flowtype and sensor from input file names, but we
     * should not consider it a complete failure */
    sksiteConfigure(0);

    /* set final delimeter */
    if (!app_flags.no_final_delimiter) {
        final_delim[0] = delimiter;
    }

    /* create the ascii stream and set its properties */
    if (rwAsciiStreamCreate(&ascii_str)) {
        skAppPrintErr("Unable to create ascii stream");
        exit(EXIT_FAILURE);
    }
    rwAsciiSetNoNewline(ascii_str);
    rwAsciiSetNoFinalDelimiter(ascii_str);
    rwAsciiSetDelimiter(ascii_str, delimiter);
    rwAsciiSetIPv6Policy(ascii_str, ipv6_policy);
    rwAsciiSetTimestampFlags(ascii_str, time_flags);
    if (app_flags.no_titles) {
        rwAsciiSetNoTitles(ascii_str);
    }
    if (app_flags.no_columns) {
        rwAsciiSetNoColumns(ascii_str);
    }
    if (app_flags.integer_sensors) {
        rwAsciiSetIntegerSensors(ascii_str);
    }

    /* conflicting properties */
    if (app_flags.zero_pad_ips && !app_flags.integer_ips) {
        rwAsciiSetZeroPadIps(ascii_str);
    }
    if (app_flags.integer_ips) {
        if (app_flags.zero_pad_ips) {
            skAppPrintErr("--integer-ips option overrides"
                          " --zero-pad-ips\n"
                          "\tWill print IPs as integers.");
        }
        rwAsciiSetIntegerIps(ascii_str);
    }

    /* handle the topn case */
    if (!proto_stats) {
        topnSetup(&leg);
    }

    /* Use STDIN as an input stream if it is not a TTY; make certain
     * we have some input and we are either reading from STDIN or
     * using files listed the command line, but not both. */
    if (iochecksAcceptFromStdin(ioISP)
        || iochecksInputs(ioISP, 0))
    {
        skAppUsage();
    }

    /* final check. See if stdout is being used for both --copy-input
     * and as the destination for the uniq data.  */
    if (ioISP->stdoutUsed && output.fp == stdout) {
        skAppPrintErr("stdout used for both --copy-input and ascii output");
        appExit(EXIT_FAILURE);
    }

    if (atexit(appTeardown) < 0) {
        skAppPrintErr("Unable to register appTeardown() with atexit()");
        appExit(EXIT_FAILURE);
    }

    /* invoke the pager */
    using_pager = skOpenPagerWhenStdoutTty(&(output.fp), &pager);
    if (using_pager < 0) {
        appExit(EXIT_FAILURE);
    }
    if (using_pager) {
        output.pathname = pager;
        output.ispipe = 1;
    }

    /* bind the Ascii Stream to the output */
    rwAsciiSetOutputHandle(ascii_str, output.fp);

    /* looks good, open the --copy-input destination */
    if (iochecksOpenCopyDest(ioISP)) {
        appExit(EXIT_FAILURE);
    }

    /* set signal handler to clean up temp files on SIGINT, SIGTERM, etc */
    if (appSetSignalHandler(&appHandleSignal)) {
        appExit(EXIT_FAILURE);
    }

    return;                       /* OK */
}


/*
 *  appTeardown()
 *
 *    Teardown all modules, close all files, and tidy up all
 *    application state.
 *
 *    This function is idempotent.
 */
void appTeardown(void)
{
    static int teardownFlag = 0;

    if (teardownFlag) {
        return;
    }
    teardownFlag = 1;

    iochecksTeardown(ioISP);

    /* plugin teardown */
    skPluginRunCleanup(SKPLUGIN_FN_ANY);
    skPluginTeardown();

    /* destroy output */
    rwAsciiStreamDestroy(&ascii_str);

    /* close output */
    if (output.fp != stdout) {
        if (output.ispipe) {
            pclose(output.fp);
        } else {
            if (EOF == fclose(output.fp) && !caught_signal) {
                skAppPrintErr("Error closing output file '%s': %s",
                              output.pathname, strerror(errno));
            }
        }
    }

    /* destroy field map and plugin-value-map */
    if (key_field_map) {
        skStringMapDestroy(key_field_map);
        key_field_map = NULL;
    }
    if (value_field_map) {
        skStringMapDestroy(value_field_map);
        value_field_map = NULL;
    }

    /* Destroy initial values array */
    if (value_initial_vals) {
        free(value_initial_vals);
        value_initial_vals = NULL;
    }

    /* destroy key */
    if (key_fields) {
        free(key_fields);
        key_fields = NULL;
    }

    /* destroy value descriptions*/
    if (value_fields) {
        free(value_fields);
        value_fields = NULL;
    }

    /* remove any temporary files */
    tempFileRemoveAll();

    skAppUnregister();
}


/*
 *  status = appOptionsHandler(cData, opt_index, opt_arg);
 *
 *    Called by skOptionsParse(), this handles a user-specified switch
 *    that the application has registered, typically by setting global
 *    variables.  Returns 1 if the switch processing failed or 0 if it
 *    succeeded.  Returning a non-zero from from the handler causes
 *    skOptionsParse() to return a negative value.
 *
 *    The clientData in 'cData' is typically ignored; 'opt_index' is
 *    the index number that was specified as the last value for each
 *    struct option in appOptions[]; 'opt_arg' is the user's argument
 *    to the switch for options that have a REQUIRED_ARG or an
 *    OPTIONAL_ARG.
 */
int appOptionsHandler(
    clientData  UNUSED(cData),
    int         opt_index,
    char       *opt_arg)
{
    static int saw_direction = 0;
    int rv;

    switch ((appOptionsEnum)opt_index) {
      case OPT_FIELDS:
        if (fields_arg) {
            skAppPrintErr("The --%s switch was specified multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        fields_arg = opt_arg;
        break;

      case OPT_VALUES:
        if (values_arg) {
            skAppPrintErr("The --%s switch was specified multiple times",
                          appOptions[opt_index].name);
            return 1;
        }
        values_arg = opt_arg;
        break;

      case OPT_TOP:
      case OPT_BOTTOM:
        if (saw_direction) {
            skAppPrintErr("May only specify one of --%s or --%s.",
                          appOptions[OPT_TOP].name,
                          appOptions[OPT_BOTTOM].name);
            return 1;
        }
        saw_direction = 1;
        if (OPT_TOP == opt_index) {
            direction = RWSTATS_DIR_TOP;
        } else {
            direction = RWSTATS_DIR_BOTTOM;
        }
        break;

      case OPT_COUNT:
      case OPT_THRESHOLD:
      case OPT_PERCENTAGE:
        if (limit.seen != 0) {
            skAppPrintErr("May only specify one of --%s, --%s, or --%s.",
                          appOptions[OPT_COUNT].name,
                          appOptions[OPT_THRESHOLD].name,
                          appOptions[OPT_PERCENTAGE].name);
            return 1;
        }
        limit.type = (opt_index - OPT_COUNT);
        rv = skStringParseUint64(&limit.value[limit.type], opt_arg,
                                 1, ((OPT_PERCENTAGE == opt_index) ? 100 : 0));
        if (rv) {
            goto PARSE_ERROR;
        }
        limit.seen = 1;
        break;

      case OPT_OVERALL_STATS:
        /* combined stats for all protocols */
        proto_stats = 1;
        break;

      case OPT_DETAIL_PROTO_STATS:
        /* detailed stats for specific proto */
        if (0 != protoStatsParse(opt_arg)) {
            return 1;
        }
        proto_stats = 1;
        break;

      case OPT_PLUGIN:
        /* plugin */
        if (skPluginLoadPluginOrDynlib(opt_arg, DYNLIB_UNIQ, 1) != 0) {
            skAppPrintErr("Unable to load %s as a plugin", opt_arg);
            return 1;
        }
        break;

      case OPT_BIN_TIME:
        if (opt_arg == NULL || opt_arg[0] == '\0') {
            /* no time given; use default */
            time_bin_size = DEFAULT_TIME_BIN;
        } else {
            /* parse user's time */
            rv = skStringParseUint32(&time_bin_size, opt_arg, 1, 0);
            if (rv) {
                goto PARSE_ERROR;
            }
        }
        break;

      case OPT_PRESORTED_INPUT:
        app_flags.presorted_input = 1;
        break;

      case OPT_NO_PERCENTS:
        app_flags.no_percents = 1;
        break;

      case OPT_EPOCH_TIME:
        if (time_flags & SKTIMESTAMP_MMDDYYYY) {
            skAppPrintErr("Both --%s and --%s specified. Using --%s",
                          appOptions[OPT_EPOCH_TIME].name,
                          appOptions[OPT_LEGACY_TIMESTAMPS].name,
                          appOptions[OPT_EPOCH_TIME].name);
        }
        time_flags = (time_flags & ~SKTIMESTAMP_MMDDYYYY) | SKTIMESTAMP_EPOCH;
        break;

      case OPT_INTEGER_IPS:
        app_flags.integer_ips = 1;
        break;

      case OPT_ZERO_PAD_IPS:
        app_flags.zero_pad_ips = 1;
        break;

      case OPT_INTEGER_SENSORS:
        app_flags.integer_sensors = 1;
        break;

      case OPT_NO_TITLES:
        app_flags.no_titles = 1;
        break;

      case OPT_NO_COLUMNS:
        memset(width, 0, RWSTATS_COLUMN_WIDTH_COUNT * sizeof(width[0]));
        app_flags.no_columns = 1;
        break;

      case OPT_NO_FINAL_DELIMITER:
        app_flags.no_final_delimiter = 1;
        break;

      case OPT_COLUMN_SEPARATOR:
        delimiter = opt_arg[0];
        break;

      case OPT_DELIMITED:
        memset(width, 0, RWSTATS_COLUMN_WIDTH_COUNT * sizeof(width[0]));
        app_flags.no_columns = 1;
        app_flags.no_final_delimiter = 1;
        if (opt_arg) {
            delimiter = opt_arg[0];
        }
        break;

      case OPT_PRINT_FILENAMES:
        app_flags.print_filenames = 1;
        break;

      case OPT_COPY_INPUT:
        if (iochecksAllDestinations(ioISP, opt_arg)) {
            return 1;
        }
        break;

      case OPT_OUTPUT_PATH:
        output.pathname = opt_arg;
        if (skOpenFile(output.pathname, 1 /* write */, &output.fp,
                       &output.ispipe))
        {
            return 1;
        }
        break;

      case OPT_PAGER:
        pager = opt_arg;
        break;

      case OPT_TEMP_DIRECTORY:
        assert(opt_arg);
        tempFileSetTempdir(opt_arg);
        break;

      case OPT_LEGACY_TIMESTAMPS:
        if ((opt_arg == NULL) || (opt_arg[0] == '\0') || (opt_arg[0] == '1')) {
            if (time_flags & SKTIMESTAMP_EPOCH) {
                skAppPrintErr("Both --%s and --%s specified; using --%s",
                              appOptions[OPT_EPOCH_TIME].name,
                              appOptions[OPT_LEGACY_TIMESTAMPS].name,
                              appOptions[OPT_EPOCH_TIME].name);
                /* no change */
            } else {
                time_flags |= SKTIMESTAMP_MMDDYYYY;
            }
        } else if (opt_arg[0] != '0') {
            skAppPrintErr("Invalid %s '%s': Select 0 or 1",
                          appOptions[opt_index].name, opt_arg);
            return 1;
        }
        break;

      case OPT_LEGACY_HELP:
        legacy_help_requested = 1;
        appUsageLong();
        exit(EXIT_SUCCESS);
    }

    return 0;                     /* OK */

  PARSE_ERROR:
    skAppPrintErr("Invalid %s '%s': %s",
                  appOptions[opt_index].name, opt_arg,
                  skStringParseStrerror(rv));
    return 1;
}


/*
 *  appExit(status)
 *
 *  Exit the application with the given status.
 */
void appExit(int status)
{
    appTeardown();
    exit(status);
}


/*
 *  appHandleSignal(signal_value)
 *
 *    Call appExit() to exit the program.  If signal_value is SIGPIPE,
 *    close cleanly; otherwise print a message that we've caught the
 *    signal and exit with EXIT_FAILURE.
 */
static void appHandleSignal(int sig)
{
    caught_signal = 1;

    if (sig == SIGPIPE) {
        /* we get SIGPIPE if something downstream, like rwcut, exits
         * early, so don't bother to print a warning, and exit
         * successfully */
        appExit(EXIT_SUCCESS);
    } else {
        skAppPrintErr("Caught signal..cleaning up and exiting");
        appExit(EXIT_FAILURE);
    }
}


/*
 *  topnSetup(legacy_args);
 *
 *    Do everything needed to setup a top-N/bottom-N run.  This
 *    function exits on error.
 */
static void topnSetup(const rwstats_legacy_t *leg)
{
    /* verify that we have an N for our top-N */
    if (!limit.seen) {
        skAppPrintErr(("No stopping condition was entered.\n"
                       "\tChoose one of --%s, --%s, or --%s"),
                      appOptions[OPT_COUNT].name,
                      appOptions[OPT_THRESHOLD].name,
                      appOptions[OPT_PERCENTAGE].name);
        skAppUsage();
    }

    /* verify that the temp directory is valid */
    if (tempFileCheckTempdir()) {
        appExit(EXIT_FAILURE);
    }

    /* set up the key_field_map and value_field_map */
    if (createStringmaps()) {
        appExit(EXIT_FAILURE);
    }

    /* make sure the user specified the --fields switch */
    if (leg->fields) {
        if (fields_arg) {
            skAppPrintErr("Cannot use --%s and old style switches",
                          appOptions[OPT_FIELDS].name);
            skAppUsage();
        }
        fields_arg = leg->fields;
    } else if (fields_arg == NULL || fields_arg[0] == '\0') {
        skAppPrintErr("The --%s switch is required",
                      appOptions[OPT_FIELDS].name);
        skAppUsage();         /* never returns */
    }

    /* find the values to use, or default to counting flows */
    if (leg->values) {
        if (values_arg != NULL) {
            skAppPrintErr("Cannot use --%s and old style switches",
                          appOptions[OPT_VALUES].name);
            skAppUsage();
        }
        values_arg = leg->values;
    } else if (values_arg == NULL) {
        values_arg = "records";
    } else if (values_arg[0] == '\0') {
        skAppPrintErr("The --%s switch is required",
                      appOptions[OPT_VALUES].name);
        skAppUsage();         /* never returns */
    }

    /* parse the --fields and --values switches */
    if (parseKeyFields(fields_arg)) {
        appExit(EXIT_FAILURE);
    }
    if (parseValueFields(values_arg)) {
        appExit(EXIT_FAILURE);
    }

    /* size of key in the hash is the key and any distinct fields */
    hash_key_bytes = key_byte_width + distinct_bytes_key;

    /* size of the nodes in the hash table */
    hash_node_bytes = hash_key_bytes + hash_value_bytes;

    /* size of the nodes in the heap */
    heap_node_bytes = key_byte_width + heap_value_bytes;

#if SK_ENABLE_IPV6
    /* we use IPSets to count distinct sIP and dIP for presorted input
     * and to count multiple distinct columns for the hash
     * case. IPSets cannot hold IPv6 addresses, so complain on these
     * conditions */
    if (distinct_num_fields && ipv6_policy >= SK_IPV6POLICY_MIX) {
        if (app_flags.presorted_input) {
            skAppPrintErr(("Distinct IPs are not yet supported for"
                           " presorted input and IPv6.\n"
                           "\tSpecify --ipv6-policy=asv4 to ignore IPv6"
                           " addresses in the input"));
            appExit(EXIT_FAILURE);
        }
        /* else we are using the hashlib */
        if (distinct_num_fields > 1) {
            skAppPrintErr(("Only a single distinct IP field is supported for"
                           " IPv6 and unsorted\n"
                           "\tinput.  Specify either a single distinct"
                           " columns or --ipv6-policy=asv4\n"
                           "\tto ignore IPv6 addresses"));
            appExit(EXIT_FAILURE);
        }
    }
#endif  /* SK_ENABLE_IPV6 */

    /* we may be restricted in the type of limit we can use depending
     * on the input and the field we're using as the key */
    if (app_flags.presorted_input) {
        if (RWSTATS_COUNT != limit.type) {
            skAppPrintErr(("Only the --%s limit is supported when --%s"
                           " is active"),
                          appOptions[OPT_COUNT].name,
                          appOptions[OPT_PRESORTED_INPUT].name);
            appExit(EXIT_FAILURE);
        }
    }
    switch (value_fields[0].vf_id) {
      case VALUE_BYTES:
      case VALUE_PACKETS:
      case VALUE_FLOWS:
        /* these are fine */
        break;

      case DISTINCT_SIP:
      case DISTINCT_DIP:
        /* count and threshold are supported */
        if (RWSTATS_PERCENTAGE == limit.type) {
            skAppPrintErr(("The %s value-field does not support --%s;\n"
                           "\tuse either --%s or --%s instead"),
                          value_fields[0].vf_title,
                          appOptions[OPT_PERCENTAGE].name,
                          appOptions[OPT_COUNT].name,
                          appOptions[OPT_THRESHOLD].name);
            appExit(EXIT_FAILURE);
        }
        break;

      case VALUE_PLUGIN:
        if (RWSTATS_COUNT != limit.type) {
            skAppPrintErr(("The %s value-field only supports using --%s"
                           " to set the limit"),
                          value_fields[0].vf_title,appOptions[OPT_COUNT].name);
            appExit(EXIT_FAILURE);
        }
        break;
    }
}


/*
 *  ok = parseValueFields();
 *
 *    Loop over all the entries in the value_field[] array; for those
 *    value fields the user wants to count, parse the parameter to
 *    the switch.  The parameter should be NULL to use the default
 *    limits, a single value ("10") or a single value followed by a
 *    single hyphen ("10-") to be used as the minimum threshold, or
 *    the minimum and maximum limits separated by a hyphen ("10-20").
 *
 *    If the user didn't provide a value field, default to counting
 *    flows.
 *
 *    Figure out all the offsets into the values byte-array for each
 *    of the value fields and the overall length of the byte-array.
 *
 *    Returns 0 on success, or 1 if the opt_arg could not be parsed.
 */
static int parseValueFields(const char *value_string)
{
    sk_vector_t *parsed_vec = NULL;
    sk_stringmap_entry_t **entry;
    uint32_t field_count;

    /* return value; assume failure */
    int rv = -1;

    /* error message generated when parsing fields */
    char *errmsg;

    distinct_field_t *d_field;
    value_field_t *v_field;
    size_t len;

    uint32_t bad_key;
    size_t i, j;
    skplugin_err_t err;

    skplugin_field_t *field_handle;

    /* vector to hold the parsed fields */
    parsed_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
    if (NULL == parsed_vec) {
        skAppPrintErr("Out of memory");
        goto END;
    }

    /* parse the field-list */
    if (skStringMapParse(value_field_map, value_string,
                         SKSTRINGMAP_DUPES_ERROR, parsed_vec, &errmsg))
    {
        skAppPrintErr("Invalid %s: %s",
                      appOptions[OPT_VALUES].name, errmsg);
        goto END;
    }

    field_count = skVectorGetCount(parsed_vec);

    /* set the number of value-fields and the value-byte-width to zero.
     * create the array to hold the information about each
     * value-field. */
    value_byte_width = 0;
    value_num_fields = 0;
    value_fields = calloc(field_count, sizeof(value_field_t));
    if (value_fields == NULL) {
        skAppPrintErr("Out of memory");
        goto END;
    }

    /* Loop over the selected values */
    for (i = 0; i < field_count; ++i) {
        v_field = &value_fields[value_num_fields];

        entry = skVectorGetValuePointer(parsed_vec, i);
        switch ((*entry)->id) {
          case VALUE_BYTES:
          case VALUE_PACKETS:
          case VALUE_FLOWS:
            assert(potential_values[(*entry)->id].vf_id == (*entry)->id);
            memcpy(v_field, &potential_values[(*entry)->id],
                   sizeof(value_field_t));

            v_field->vf_offset = value_byte_width;
            value_byte_width += v_field->vf_byte_width;
            ++value_num_fields;
            break;

          case DISTINCT_SIP:
          case DISTINCT_DIP:
            /* Counting distinct IPs when that IP is part of the key
             * doesn't make sense (the count is always 1). */
            if ((*entry)->id == DISTINCT_SIP) {
                bad_key = RWREC_FIELD_SIP;
                errmsg = "sIP";
            } else {
                bad_key = RWREC_FIELD_DIP;
                errmsg = "dIP";
            }
            for (j = 0; j < key_num_fields; ++j) {
                if (bad_key == key_fields[j].kf_id) {
                    skAppPrintErr(("Will not count distinct %ss"
                                   " when %s is part of key"),
                                  errmsg, errmsg);
                    goto END;
                }
            }
            /* the distinct fields are part of the key in the hash
             * table, but are printed as values.  add to value_fields,
             * but do not account for their byte-width yet, and only
             * handle offset relative to other distinct fields. */
            assert(potential_values[(*entry)->id].vf_id == (*entry)->id);
            memcpy(v_field, &potential_values[(*entry)->id],
                   sizeof(value_field_t));
            v_field->vf_offset = distinct_bytes_value;
            distinct_bytes_value += v_field->vf_byte_width;

            /* keep track of how the hash key needs to change */
            d_field = &distinct_fields[distinct_num_fields];
            memset(d_field, 0, sizeof(distinct_field_t));

            d_field->df_val_field_pos = value_num_fields;
            d_field->df_id = (*entry)->id;
            d_field->df_offset = key_byte_width + distinct_bytes_key;
#if SK_ENABLE_IPV6
            if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                if ((*entry)->id == DISTINCT_SIP) {
                    d_field->df_r2k = rwrec_MemGetSIPv6;
                } else {
                    d_field->df_r2k = rwrec_MemGetDIPv6;
                }
                d_field->df_byte_width = RWREC_SIZEOF_SIPv6;
            } else
#endif
            {
                if ((*entry)->id == DISTINCT_SIP) {
                    d_field->df_r2k = rwrec_MemGetSIPv4;
                } else {
                    d_field->df_r2k = rwrec_MemGetDIPv4;
                }
                d_field->df_byte_width = RWREC_SIZEOF_SIPv4;
            }
            distinct_bytes_key += d_field->df_byte_width;
            ++distinct_num_fields;
            ++value_num_fields;
            break;

          default:
            /* field comes from a plug-in */
            field_handle = (skplugin_field_t*)(*entry)->userdata;
            assert(field_handle);

            v_field->vf_id = VALUE_PLUGIN;
            v_field->vf_plugin_field = field_handle;
            err = skPluginFieldTitle(field_handle, &v_field->vf_title);
            assert(err == SKPLUGIN_OK);
            err = skPluginFieldGetLenBin(field_handle, &v_field->vf_byte_width);
            assert(err == SKPLUGIN_OK);
            /* use itermediate variable due to type differences */
            err = skPluginFieldGetLenText(field_handle, &len);
            v_field->vf_text_len = (int)len;
            assert(err == SKPLUGIN_OK);

            v_field->vf_offset = value_byte_width;
            value_byte_width += v_field->vf_byte_width;
            ++value_num_fields;
            break;
        }
    }

    /* adjust offset of distinct fields in the value_fields array */
    for (i = 0, v_field = value_fields; i < value_num_fields; ++i, ++v_field) {
        switch (v_field->vf_id) {
          case DISTINCT_SIP:
          case DISTINCT_DIP:
            v_field->vf_offset += value_byte_width;
            break;
          default:
            break;
        }
    }

    heap_value_bytes = value_byte_width + distinct_bytes_value;

    if (heap_value_bytes > MAX_VALUE_BYTE_WIDTH) {
        skAppPrintErr("The selected values exceeds the maximum value "
                      "byte width");
        goto END;
    }

    if (0 == value_byte_width) {
        /* if only distinct counts were requested, there will no value
         * fields.  however, the hash table does not accept a value
         * byte size of 0, so use a single byte.  Also, we must
         * allocate the initial value array  */
        hash_value_bytes = 1;

        value_initial_vals = calloc(hash_value_bytes, sizeof(uint8_t));
        if (NULL == value_initial_vals) {
            skAppPrintErr("Memory error default initial values");
            goto END;
        }
    } else {
        hash_value_bytes = value_byte_width;

        /* allocate a byte-array to hold the initial value for each
         * aggregate field */
        value_initial_vals = calloc(value_byte_width, sizeof(uint8_t));
        if (NULL == value_initial_vals) {
            skAppPrintErr("Memory error default initial values");
            goto END;
        }

        /* handle any plug-in fields in that byte-array */
        for (i = 0; i < value_num_fields; i++) {
            if (value_fields[i].vf_plugin_field) {
                err = skPluginFieldGetInitialValue(
                    value_fields[i].vf_plugin_field,
                    value_initial_vals + value_fields[i].vf_offset);
                if (err != SKPLUGIN_OK) {
                    skAppPrintErr("Error initializing plugin value");
                    appExit(EXIT_FAILURE);
                }
            }
        }
    }

    rv = 0;

  END:
    /* do standard clean-up */
    if (parsed_vec != NULL) {
        skVectorDestroy(parsed_vec);
    }
    if (rv != 0) {
        /* something went wrong. do additional clean-up */
        if (value_fields) {
            free(value_fields);
            value_fields = NULL;
        }
        if (value_initial_vals) {
            free(value_initial_vals);
            value_initial_vals = NULL;
        }
    }

    return rv;
}


/*
 *  status = parseKeyFields(field_string);
 *
 *    Parse the string that represents the columns the user wishes to
 *    "stats" over, fill in the key_fields[] array, and add columns to
 *    the rwAsciiStream.  Return 0 on success or non-zero on error.
 */
static int parseKeyFields(const char *field_string)
{
    sk_vector_t *parsed_vec = NULL;
    sk_stringmap_entry_t **entry;
    uint32_t field_count;

    /* return value; assume failure */
    int rv = 1;

    /* error message generated when parsing fields */
    char *errmsg;

    /* whether we've added the dPort key.  Both DPORT and
     * ICMP_TYPE_CODE map to the dPort. */
    unsigned int saw_dport = 0;

    /* which of class and type were seen */
    unsigned int saw_class_type = 0;

    uint32_t i;

    /* vector to hold the parsed fields */
    parsed_vec = skVectorNew(sizeof(sk_stringmap_entry_t*));
    if (NULL == parsed_vec) {
        skAppPrintErr("Out of memory");
        goto END;
    }

    /* parse the field-list */
    if (skStringMapParse(key_field_map, field_string, SKSTRINGMAP_DUPES_ERROR,
                         parsed_vec, &errmsg))
    {
        skAppPrintErr("Invalid %s: %s",
                      appOptions[OPT_FIELDS].name, errmsg);
        goto END;
    }

    field_count = skVectorGetCount(parsed_vec);

    /* set the number of key-fields and the key-byte-width to zero.
     * create the array to hold the information about each
     * key-field. */
    key_byte_width = 0;
    key_num_fields = 0;
    key_fields = calloc(field_count, sizeof(key_field_t));
    if (key_fields == NULL) {
        skAppPrintErr("Out of memory");
        goto END;
    }

    /* Go over the list of fields.  For "simple" fields that don't
     * require any special processing, add them to the key.  Anything
     * that we have to handle specially will wait until the second
     * pass.  For example, how we hande eTime depends on whether sTime
     * and/or duration are also specified. */
    for (i = 0; i < field_count; ++i) {
        entry = skVectorGetValuePointer(parsed_vec, i);
        if ((*entry)->userdata) {
            /* value comes from a plugin. handle it later */
            continue;
        }
        switch ((rwrec_printable_fields_t)(*entry)->id) {
          case RWREC_FIELD_STIME:
          case RWREC_FIELD_STIME_MSEC:
            time_fields |= PARSE_KEY_STIME;
            break;

          case RWREC_FIELD_ELAPSED:
          case RWREC_FIELD_ELAPSED_MSEC:
            time_fields |= PARSE_KEY_ELAPSED;
            break;

          case RWREC_FIELD_ETIME:
          case RWREC_FIELD_ETIME_MSEC:
            time_fields |= PARSE_KEY_ETIME;
            break;

          case RWREC_FIELD_SIP:
#if SK_ENABLE_IPV6
            if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                addKeyField((*entry)->id, RWREC_SIZEOF_SIPv6, NULL, NULL);
                break;
            }
#endif  /* SK_ENABLE_IPV6 */
            addKeyField((*entry)->id, RWREC_SIZEOF_SIPv4, NULL, NULL);
            break;

          case RWREC_FIELD_DIP:
#if SK_ENABLE_IPV6
            if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                addKeyField((*entry)->id, RWREC_SIZEOF_DIPv6, NULL, NULL);
                break;
            }
#endif /* SK_ENABLE_IPV6 */
            addKeyField((*entry)->id, RWREC_SIZEOF_DIPv4, NULL, NULL);
            break;

          case RWREC_FIELD_NHIP:
#if SK_ENABLE_IPV6
            if (ipv6_policy >= SK_IPV6POLICY_MIX) {
                addKeyField((*entry)->id, RWREC_SIZEOF_NHIPv6, NULL, NULL);
                break;
            }
#endif /* SK_ENABLE_IPV6 */
            addKeyField((*entry)->id, RWREC_SIZEOF_NHIPv4, NULL, NULL);
            break;

          case RWREC_FIELD_SPORT:
            addKeyField((*entry)->id, RWREC_SIZEOF_SPORT, NULL, NULL);
            break;

          case RWREC_FIELD_DPORT:
          case RWREC_FIELD_ICMP_TYPE_CODE:
            if (0 == saw_dport) {
                saw_dport = 1;
                addKeyField((*entry)->id, RWREC_SIZEOF_DPORT, NULL, NULL);
            }
            break;

          case RWREC_FIELD_PROTO:
            addKeyField((*entry)->id, RWREC_SIZEOF_PROTO, NULL, NULL);
            break;

          case RWREC_FIELD_PKTS:
            addKeyField((*entry)->id, RWREC_SIZEOF_PKTS, NULL, NULL);
            break;

          case RWREC_FIELD_BYTES:
            addKeyField((*entry)->id, RWREC_SIZEOF_BYTES, NULL, NULL);
            break;

          case RWREC_FIELD_FLAGS:
            addKeyField((*entry)->id, RWREC_SIZEOF_FLAGS, NULL, NULL);
            break;

          case RWREC_FIELD_SID:
            addKeyField((*entry)->id, RWREC_SIZEOF_SID, NULL, NULL);
            break;

          case RWREC_FIELD_INPUT:
            addKeyField((*entry)->id, RWREC_SIZEOF_INPUT, NULL, NULL);
            break;

          case RWREC_FIELD_OUTPUT:
            addKeyField((*entry)->id, RWREC_SIZEOF_OUTPUT, NULL, NULL);
            break;

          case RWREC_FIELD_INIT_FLAGS:
            addKeyField((*entry)->id, RWREC_SIZEOF_INIT_FLAGS, NULL, NULL);
            break;

          case RWREC_FIELD_REST_FLAGS:
            addKeyField((*entry)->id, RWREC_SIZEOF_REST_FLAGS, NULL, NULL);
            break;

          case RWREC_FIELD_TCP_STATE:
            addKeyField((*entry)->id, RWREC_SIZEOF_TCP_STATE, NULL, NULL);
            break;

          case RWREC_FIELD_APPLICATION:
            addKeyField((*entry)->id, RWREC_SIZEOF_APPLICATION, NULL, NULL);
            break;

          case RWREC_FIELD_FTYPE_CLASS:
          case RWREC_FIELD_FTYPE_TYPE:
            if (0 == saw_class_type) {
                saw_class_type = 1;
                addKeyField((*entry)->id, RWREC_SIZEOF_FLOW_TYPE, NULL, NULL);
            }
            break;
        }
    }

    /* Warn if --bin-time requested but time not part of key.  Warn
     * that duration field will be modified when all time fields are
     * given. */
    if (time_bin_size != 0) {
        if (time_fields < 2) {
            if (FILEIsATty(stderr)) {
                skAppPrintErr("No time fields in key; %s switch ignored",
                              appOptions[OPT_BIN_TIME].name);
            }
        } else if (time_fields == 7) {
            /* must adjust elapsed to be eTime-sTime */
            adjust_elapsed = 1;
            if (FILEIsATty(stderr)) {
                skAppPrintErr("Warning: modifying duration field "
                              "to be eTime-sTime");
            }
        }
    }

    /* Handle the time fields.  Note that the eTime field must be last
     * since it is computed from the sTime and elapsed fields. */
    if (time_fields & PARSE_KEY_STIME) {
        /* use size of 4 since we convert the sktime_t to seconds
         * since the UNIX epoch */
        addKeyField(RWREC_FIELD_STIME, 4, &recToKeyStime, &keyToRecStime);
    }
    if (time_fields & PARSE_KEY_ELAPSED) {
        addKeyField(RWREC_FIELD_ELAPSED, RWREC_SIZEOF_ELAPSED,
                    &recToKeyElapsed, &keyToRecElapsed);
    }
    if (time_fields & PARSE_KEY_ETIME) {
        addKeyField(RWREC_FIELD_ETIME, 4, &recToKeyEtime, &keyToRecEtime);
    }

    /* Go over the fields again, setting up the 'ascii_str' output and
     * filling in the key_fields[] for the plugins. */
    for (i = 0; i < field_count; ++i) {
        entry = skVectorGetValuePointer(parsed_vec, i);
        if (NULL == (*entry)->userdata) {
            /* field is built-in */
            if (rwAsciiAppendOneField(ascii_str, (*entry)->id)) {
                skAppPrintErr("Cannot add field %s to stream",
                              (*entry)->name);
                goto END;
            }
        } else {
            /* field comes from a plug-in */
            assert((*entry)->id >= RWREC_PRINTABLE_FIELD_COUNT);

            if (appAddPluginKey(*entry)) {
                skAppPrintErr("Cannot add field %s from plugin",
                              (*entry)->name);
                goto END;
            }
        }
    }

    /* check size of key */
    if (key_byte_width > MAX_KEY_BYTE_WIDTH) {
        skAppPrintErr("Fields make too large a key: %u bytes > %u max",
                      (unsigned int)key_byte_width, MAX_KEY_BYTE_WIDTH);
        goto END;
    }

    /* successful */
    rv = 0;

  END:
    if (rv != 0) {
        /* something went wrong.  clean up */
        if (key_fields) {
            free(key_fields);
            key_fields = NULL;
        }
    }
    /* do standard clean-up */
    if (parsed_vec != NULL) {
        skVectorDestroy(parsed_vec);
    }

    return rv;
}


/*
 *  appPluginGetTitle(buf, buf_size, keyfield);
 *
 *    Invoke the appropriate function on the plugin's field-handle,
 *    referenced by 'keyfield', to get the title of the field, and put
 *    that value into the buffer 'buf'.
 */
static void appPluginGetTitle(
    char        *text_buf,
    size_t       text_buf_size,
    void        *cb_data)
{
    key_field_t *keyfield = (key_field_t*)cb_data;
    skplugin_err_t err;
    const char *title;

    err = skPluginFieldTitle(keyfield->kf_field_handle, &title);
    assert(SKPLUGIN_OK == err);
    strncpy(text_buf, title, text_buf_size);
    text_buf[text_buf_size-1] = '\0';
}


/*
 *  appPluginGetTitle(rwrec, buf, buf_size, keyfield);
 *
 *    Invoke the appropriate function on the plugin's field-handle,
 *    referenced by 'keyfield', to get the value of the field, and put
 *    that value into the buffer 'buf'.
 */
static int appPluginGetValue(
    const rwRec UNUSED(*rwrec),
    char               *text_buf,
    size_t              text_buf_size,
    void               *cb_data)
{
    key_field_t *keyfield = (key_field_t*)cb_data;

    /* The text buffer exists on the keyfield and it has already been
     * filled, so just copy that value into 'buf'.  Yes, this is kind
     * of pointless. */
    strncpy(text_buf, keyfield->kf_textbuf, text_buf_size);
    return 0;
}

/*
 *  status = appAddPluginKey(map_entry);
 *
 *    Given an entry from the key_field_map that represents a key field
 *    from a plug-in, activate that field and get the information from
 *    the field that rwstats requires.  Add that key to the global
 *    key_field[] array and update the global 'key_byte_width' and
 *    'key_num_fields' values.  Finally, add that field to the
 *    rwAsciiStream.
 */
static int appAddPluginKey(
    sk_stringmap_entry_t   *map_entry)
{
    key_field_t      *keyfield;
    skplugin_field_t *field_handle;
    size_t            bin_width;
    size_t            text_width;
    skplugin_err_t    err;

    field_handle = (skplugin_field_t*)map_entry->userdata;

    /* Activate the field (so cleanup knows about it) */
    err = skPluginFieldActivate(field_handle);
    if (err != SKPLUGIN_OK) {
        return -1;
    }

    /* Initialize this field */
    err = skPluginFieldRunInitialize(field_handle);
    if (err != SKPLUGIN_OK) {
        return -1;
    }

    /* get the required width of the column; but don't allow a value
     * larger than the maximum column width */
    err = skPluginFieldGetLenText(field_handle, &text_width);
    if (err != SKPLUGIN_OK) {
        return -1;
    }
    if (text_width == 0 || text_width >= VALUE_OUT_BUF_SIZE) {
        text_width = VALUE_OUT_BUF_SIZE - 1;
    }

    /* get the bin width for this field */
    err = skPluginFieldGetLenBin(field_handle, &bin_width);
    if (err != SKPLUGIN_OK) {
        return -1;
    }

    /* add a new key field */
    keyfield = &(key_fields[key_num_fields]);

    keyfield->kf_id           = map_entry->id;
    keyfield->kf_offset       = key_byte_width;
    keyfield->kf_width        = bin_width;
    keyfield->kf_field_handle = field_handle;
    keyfield->kf_r2k          = NULL;
    keyfield->kf_k2r          = NULL;

    key_byte_width += keyfield->kf_width;
    ++key_num_fields;

    return rwAsciiAppendCallbackField(ascii_str, &appPluginGetTitle,
                                      &appPluginGetValue, keyfield,
                                      text_width);
}


/*
 *  keyfield = addKeyField(field_id, bin_width, rec2key, key2rec);
 *
 *    Add a new "built-in" field to the attributes that will make up
 *    the key we use for binning, that is, to the global 'key_fields'
 *    array, adjusting the 'key_num_fields' and 'key_byte_width'
 *    globals as well.
 *
 *    'field_id' is the ID that this field has in the global key_field_map
 *    (e.g., RWREC_FIELD_SPORT for the source-port field).
 *
 *    'bin_width' is the number of bytes required to hold this field's
 *    value in the key.  For example, 2 for the source port.  This
 *    value must be non-zero.
 *
 *    The functions specified in 'rec2key' and 'key2rec' may be NULL.
 *    If they are non-NULL, they are functions used to extract a
 *    binary value from an rwRec during input; and during output it is
 *    used to convert that binary value back into field(s) on an
 *    rwRec.  They are used for fields such as eTime that don't have
 *    actual field on the rwRec.
 *
 *    This function returns a pointer into the global 'key_fields'
 *    array that is a handle to the newly added key.
 */
static key_field_t *addKeyField(
    uint32_t        field_id,
    size_t          bin_width,
    rec2key_fn_t    rec2key,
    key2rec_fn_t    key2rec)
{
    key_field_t *keyfield;

    assert(bin_width > 0);

    /* add the field */
    keyfield = &(key_fields[key_num_fields]);
    keyfield->kf_id           = field_id;
    keyfield->kf_offset       = key_byte_width;
    keyfield->kf_width        = bin_width;
    keyfield->kf_r2k          = rec2key;
    keyfield->kf_k2r          = key2rec;
    keyfield->kf_field_handle = NULL;

    /* adjust counter and byte width */
    key_byte_width += keyfield->kf_width;
    ++key_num_fields;

    return keyfield;
}


/*
 *  ok = recToKeyStime(rwrec, bin_value);
 *
 *    Copy the start-time value (in seconds) from 'rwrec' into
 *    'bin_value'.  If the user requested the floor of the start-time,
 *    do that calculation.  The start-time is always stored in network
 *    byte order.
 */
static void recToKeyStime(
    const rwRec    *rwrec,
    void           *bin_value)
{
    uint32_t sTime;

    if (time_bin_size <= 1) {
        sTime = htonl(rwRecGetStartSeconds(rwrec));
    } else {
        sTime = rwRecGetStartSeconds(rwrec);
        /* find the floor of the time */
        sTime = htonl(sTime - (sTime % time_bin_size));
    }
    memcpy(bin_value, &sTime, sizeof(sTime));
}


/*
 *  keyToRecStime(rwrec, bin_value);
 *
 *    Copy the value from 'bin_value' to the start-time field in the
 *    'rwrec'.
 */
static void keyToRecStime(
    rwRec          *rwrec,
    const void     *bin_value)
{
    uint32_t sTime;

    memcpy(&sTime, bin_value, sizeof(sTime));
    rwRecSetStartTime(rwrec, sktimeCreate(ntohl(sTime), 0));
}


/*
 *  recToKeyElapsed(rwrec, bin_value);
 *
 *    Get the elapsed field from the 'rwrec' and adjust it so that the
 *    computed start-time and end-time fields fall into the proper
 *    bins with respect to the global 'time_bin_size'.  Put that
 *    computed elapsed value into 'bin_value' in network byte order.
 */
static void recToKeyElapsed(
    const rwRec    *rwrec,
    void           *bin_value)
{
    uint32_t elapsed;
    uint32_t sTime;

    if (!adjust_elapsed) {
        elapsed = htonl(rwRecGetElapsedSeconds(rwrec));
    } else {
        /*
         * set elapsed to:
         * (eTime - (eTime % bin_size)) - (sTime - (sTime % bin_size))
         */
        sTime = rwRecGetStartSeconds(rwrec);
        elapsed = rwRecGetElapsedSeconds(rwrec);
        elapsed = htonl(elapsed + (sTime % time_bin_size)
                        - ((sTime + elapsed) % time_bin_size));
    }
    memcpy(bin_value, &elapsed, sizeof(elapsed));
}


/*
 *  keyToRecElapsed(rwrec, bin_value);
 *
 *    Copy the value from 'bin_value' to the elapsed field in the
 *    'rwrec'.
 */
static void keyToRecElapsed(
    rwRec          *rwrec,
    const void     *bin_value)
{
    uint32_t elapsed;

    memcpy(&elapsed, bin_value, sizeof(elapsed));
    rwRecSetElapsed(rwrec, 1000 * ntohl(elapsed));
}


/*
 *  recToKeyEtime(rwrec, bin_value);
 *
 *    Use the start-time and elapsed fields on the 'rwrec' to compute
 *    the end-time and put the value into 'bin_value'.  Adjust the
 *    end-time relative to the global 'time_bin_size' if required.
 */
static void recToKeyEtime(
    const rwRec    *rwrec,
    void           *bin_value)
{
    uint32_t eTime;

    /* compute eTime and put it into bin_value */
    if (time_bin_size == 0) {
        eTime = htonl(rwRecGetEndSeconds(rwrec));
    } else {
        /* compute floor */
        eTime = rwRecGetEndSeconds(rwrec);
        eTime = htonl(eTime - (eTime % time_bin_size));
    }
    memcpy(bin_value, &eTime, sizeof(eTime));
}


/*
 *  keyToRecEtime(rwrec, bin_value);
 *
 *    Adjust the start-time and elapsed fields on the 'rwrec' so that
 *    the end-time will have the value present in 'bin_value'.  This
 *    function assumes the start-time and elapsed values have already
 *    been copied from the byte-array.
 */
static void keyToRecEtime(
    rwRec          *rwrec,
    const void     *bin_value)
{
    uint32_t eTime, sTime, elapsed;

    memcpy(&eTime, bin_value, sizeof(eTime));
    eTime = ntohl(eTime);

    switch (time_fields) {
      case 4:
        /* etime only; just set sTime to eTime--elapsed is already 0 */
        rwRecSetStartTime(rwrec, sktimeCreate(eTime, 0));
        break;

      case 5:
        /* etime and elapsed; set start time based on end time and elapsed */
        elapsed = rwRecGetElapsedSeconds(rwrec);
        rwRecSetStartTime(rwrec, sktimeCreate((eTime - elapsed), 0));
        break;

      case 6:
        /* etime and stime; set elapsed as their difference */
        sTime = rwRecGetStartSeconds(rwrec);
        assert(sTime <= eTime);
        rwRecSetElapsed(rwrec, (1000 * (eTime - sTime)));
        break;

      case 7:
        /* etime, stime, elapsed; nothing to do */
        assert((rwRecGetStartSeconds(rwrec) + rwRecGetElapsedSeconds(rwrec))
               == eTime);
        break;

      default:
        abort();
    }
}


/*
 *  int = appNextInput(&rwios);
 *
 *    Fill 'rwios' with the next input file to read.  Return 0 if
 *    'rwios' was successfully opened or 1 if there are no more input
 *    files.
 *
 *    When an input file cannot be opened, the return value is
 *    dependent on the error.  If the error is due to being out of
 *    file handles or memory (EMFILE or ENOMEM), return -2; otherwise
 *    return -1.
 */
int appNextInput(skstream_t **rwios)
{
    static int counter = -1;
    int rv;

    if (counter < 0) {
        counter = ioISP->firstFile;
    } else {
        ++counter;
        if (counter == ioISP->fileCount) {
            /* no more input */
            return 1;
        }
    }

    /* create rwios and open file */
    errno = 0;
    rv = skStreamOpenSilkFlow(rwios, ioISP->fnArray[counter], SK_IO_READ);
    if (rv) {
        if (errno == EMFILE || errno == ENOMEM) {
            /* decrement counter so we try to open this file again */
            --counter;
            rv = -2;
            PRINTDEBUG((stderr, "unable to open %s because of %s\n",
                        ioISP->fnArray[counter], strerror(errno)));
        } else {
            if (app_flags.print_filenames) {
                fprintf(PRINT_FILENAMES_FH, "%s\n", ioISP->fnArray[counter]);
            }
            skStreamPrintLastErr(*rwios, rv, &skAppPrintErr);
            rv = -1;
        }
        skStreamDestroy(rwios);
        return rv;
    }

    if (app_flags.print_filenames) {
        fprintf(PRINT_FILENAMES_FH, "%s\n", skStreamGetPathname(*rwios));
    }
    skStreamSetCopyInput(*rwios, ioISP->inputCopyFD);
    skStreamSetIPv6Policy(*rwios, ipv6_policy);

    return 0;
}


/*
 *  ok = createStringmaps();
 *
 *    Create the string-maps to assist in parsing the --fields and
 *    --values switches.
 */
static int createStringmaps(void)
{
    skplugin_field_iter_t iter;
    skplugin_err_t err;
    skplugin_field_t *field_handle;
    sk_stringmap_entry_t entry;
    sk_stringmap_status_t map_err;
    const char **field_names;
    const char **name;
    size_t num_builtin_values;
    uint32_t max_id;
    size_t i;

    /*
     * initialize string-map of key field identifiers: add default
     * keys, then remove millisec fields, since using them as a key
     * makes little sense.
     */
    if (rwAsciiFieldMapAddDefaultFields(&key_field_map)) {
        skAppPrintErr("Unable to setup fields stringmap");
        return -1;
    }
    (void)skStringMapRemoveByID(key_field_map, RWREC_FIELD_STIME_MSEC);
    (void)skStringMapRemoveByID(key_field_map, RWREC_FIELD_ETIME_MSEC);
    (void)skStringMapRemoveByID(key_field_map, RWREC_FIELD_ELAPSED_MSEC);
    max_id = RWREC_PRINTABLE_FIELD_COUNT - 1;

    /* add any key fields from the plug-ins  */
    err = skPluginFieldIteratorBind(&iter, SKPLUGIN_APP_STATS_FIELD, 1);
    assert(err == SKPLUGIN_OK);
    while (skPluginFieldIteratorNext(&iter, &field_handle)) {
        err = skPluginFieldName(field_handle, &field_names);
        assert(err == SKPLUGIN_OK);
        ++max_id;

        /* Add keys to the key_field_map */
        for (name = field_names; *name; name++) {
            memset(&entry, 0, sizeof(entry));
            entry.name = *name;
            entry.id = max_id;
            entry.userdata = field_handle;
            map_err = skStringMapAddEntries(key_field_map, 1, &entry);
            if (map_err != SKSTRINGMAP_OK) {
                const char *plugin_name;
                skPluginFieldGetPluginName(field_handle, &plugin_name);
                skAppPrintErr(("Plug-in cannot add field named '%s': %s."
                               " Plug-in file: %s"),
                              *name, skStringMapStrerror(map_err),plugin_name);
                return -1;
            }
        }
    }


    max_id = 0;

    /* create the string-map for value field identifiers. */
    if (skStringMapCreate(&value_field_map)) {
        skAppPrintErr("Unable to create map for values");
        return -1;
    }

    /* add the built-in names */
    num_builtin_values = (sizeof(potential_values)/sizeof(value_field_t));
    for (i = 0; i < num_builtin_values; ++i) {
        memset(&entry, 0, sizeof(sk_stringmap_entry_t));
        entry.name = potential_values[i].vf_title;
        entry.id = potential_values[i].vf_id;

        map_err = skStringMapAddEntries(value_field_map, 1, &entry);
        if (map_err) {
            skAppPrintErr("Unable to add value field named '%s': %s",
                          entry.name, skStringMapStrerror(map_err));
            return -1;
        }
        if (entry.id > max_id) {
            max_id = entry.id;
        }
    }

    /* add the value fields from the plugins */
    err = skPluginFieldIteratorBind(&iter, SKPLUGIN_APP_STATS_VALUE, 1);
    assert(err == SKPLUGIN_OK);
    while (skPluginFieldIteratorNext(&iter, &field_handle)) {
        err = skPluginFieldName(field_handle, &field_names);
        assert(err == SKPLUGIN_OK);
        ++max_id;

        /* Add values to the value_field_map */
        for (name = field_names; *name; ++name) {
            memset(&entry, 0, sizeof(entry));
            entry.name = *name;
            entry.id = max_id;
            entry.userdata = field_handle;
            map_err = skStringMapAddEntries(value_field_map, 1, &entry);
            if (map_err != SKSTRINGMAP_OK) {
                const char *plugin_name;
                skPluginFieldGetPluginName(field_handle, &plugin_name);
                skAppPrintErr(("Plug-in cannot add value named '%s': %s."
                               " Plug-in file: %s"),
                              *name, skStringMapStrerror(map_err),plugin_name);
                return -1;
            }
        }
    }

    return 0;
}


/*
** Local Variables:
** mode:c
** indent-tabs-mode:nil
** c-basic-offset:4
** End:
*/
